
// This file is part of the Herezh++ application.
//
// The finite element software Herezh++ is dedicated to the field
// of mechanics for large transformations of solid structures.
// It is developed by Gérard Rio (APP: IDDN.FR.010.0106078.000.R.P.2006.035.20600)
// INSTITUT DE RECHERCHE DUPUY DE LÔME (IRDL) <https://www.irdl.fr/>.
//
// Herezh++ is distributed under GPL 3 license ou ultérieure.
//
// Copyright (C) 1997-2022 Université Bretagne Sud (France)
// AUTHOR : Gérard Rio
// E-MAIL  : gerardrio56@free.fr
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License,
// or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.
//
// For more information, please consult: <https://herezh.irdl.fr/>.

/************************************************************************
 *     DATE:        23/01/97                                            *
 *                                                                $     *
 *     AUTEUR:      G RIO   (mailto:gerardrio56@free.fr)                *
 *                                                                $     *
 *     PROJET:      Herezh++                                            *
 *                                                                $     *
 ************************************************************************
 *     BUT:  Def de la geometrie d'une droite: un point et un vecteur   *
 *           directeur.                                                 *
 *                                                                $     *
 *     ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''     *                                                                      *
 *     VERIFICATION:                                                    *
 *                                                                      *
 *     !  date  !   auteur   !       but                          !     *
 *     ------------------------------------------------------------     *
 *     !        !            !                                    !     *
 *                                                                $     *
 *     ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''     *
 *     MODIFICATIONS:                                                   *
 *     !  date  !   auteur   !       but                          !     *
 *     ------------------------------------------------------------     *
 *                                                                $     *
 ************************************************************************/
#ifndef DROITE_H
#define DROITE_H

#include "Coordonnee.h"

/// @addtogroup Groupe_sur_les_contacts
///  @{
///


class Droite
{   // surcharge de l'operator de lecture
    friend istream & operator >> (istream &, Droite &);
    // surcharge de l'operator d'ecriture
    friend ostream & operator << (ostream &, const Droite &);

   public :
    // CONSTRUCTEURS :
    // par defaut définit une droit // à x et passant par 0, et de dimension celle de l'espace de travail
    Droite ();
    // avec les datas
    // vec est quelconque, non norme
    Droite ( const Coordonnee& B, const Coordonnee& vec);
    // avec la dimension: la droite générée par défaut est une droite qui passe par 0 et est // à l'axe des x
    Droite (int dim);
    // de copie
    Droite ( const Droite& a);
    // DESTRUCTEUR :
    ~Droite ();
    // surcharge des operator
    Droite& operator = ( const Droite & P);
        
    // METHODES PUBLIQUES :
    // retour le  point de ref la droite 
    inline const Coordonnee& PointDroite()  const { return A;};
    // retourne le vecteur directeur de la droite (de norme = 1 )
    inline const Coordonnee& VecDroite()  const { return U;};
    
    // change la dimension de la droite
    void Change_dim(int dima) {A.Change_dim(dima);U.Change_dim(dima);};
    // change le point de ref de la droite
    void Change_ptref( const Coordonnee& B);
    // change le vecteur de la droite
    void Change_vect( const Coordonnee& vec);
    // change toutes les donnees
    void change_donnees( const Coordonnee& B, const Coordonnee& vec);
   
    // calcul l'intercection M de la droite avec une autre droite, ramene  0 s'il n'y
    // a pas d'Intersection, ramene -1 si l'Intersection ne peut pas etre calculee
    // et 1 s'il y a un point d'Intersection
    // NB: en axisymétrique, on considère que les droites sont dans le plan xy
    int Intersection( const Droite & D,Coordonnee& M);
    
    // calcul si un point donné appartient ou non à la droite (à la précision donnée)
    bool Appartient_a_la_droite(const Coordonnee& B);
    
    // ramene une normale, en 2D il y a une solution, en 3D a chaque appel on ramene
    // une nouvelle normale calculée aleatoirement
    // en 3D axi ramène la normale dans le plan xy : solution unique si elle peut-être calculée
    Coordonnee UneNormale();
    
    // calcul la distance d'un point à la droite
    double Distance_a_la_droite(const Coordonnee& M) const;
    
    // calcul du projeté d'un point sur la doite
    Coordonnee Projete(const Coordonnee& M) const 
               {return (A+((M-A)*U) * U);};
               
    // fonction valable uniquement en 2D !!!, sinon erreur
    // ramène true si les deux points sont du même coté de la droite, false sinon
    bool DuMemeCote(const Coordonnee& M1, const Coordonnee& M2) const;           
         
    
  protected :  
    // VARIABLES PROTEGEES :
    Coordonnee A; // un point de la droite
    Coordonnee U; // vecteur directeur de la droite
    static int alea; //  une variables aleatoire qui sert pour la sortie d'une normale
    // aleatoire en 3D
    // METHODES PROTEGEES :
 };
 /// @}  // end of group

#endif  
